import asyncio
import json
import logging
import os
import time

from openai import AsyncOpenAI

from realtime.plugins.base_plugin import Plugin

logger = logging.getLogger(__name__)


class OpenAILLM(Plugin):
    def __init__(
        self,
        model: str = "gpt-3.5-turbo",
        api_key=None,
        base_url=None,
        system_prompt=None,
    ):
        super().__init__()
        self._model: str = model
        api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError("OpenAI API key is required")
        self._api_key = api_key
        self._client = AsyncOpenAI(api_key=self._api_key, base_url=base_url)
        self._history = []
        self.output_queue = asyncio.Queue()
        self.chat_history_queue = asyncio.Queue()
        self._generating = False
        self._system_prompt = system_prompt
        if self._system_prompt is not None:
            self._history.append({"role": "system", "content": self._system_prompt})

    async def _stream_chat_completions(self):
        while True:
            text_chunk = await self.input_queue.get()
            if text_chunk is None:
                continue
            self._generating = True
            self._history.append({"role": "user", "content": text_chunk})
            self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
            start_time = time.time()
            chunk_stream = await self._client.chat.completions.create(
                model=self._model,
                stream=True,
                messages=self._history,
            )
            logger.info("OpenAI LLM TTFB: %s", time.time() - start_time)
            self._history.append({"role": "assistant", "content": ""})
            async for chunk in chunk_stream:
                if len(chunk.choices) == 0:
                    continue

                elif chunk.choices[0].delta.content:
                    self._history[-1]["content"] += chunk.choices[0].delta.content
                    await self.output_queue.put(chunk.choices[0].delta.content)
            self._generating = False
            self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
            await self.output_queue.put(None)

    async def run(self, input_queue: asyncio.Queue) -> asyncio.Queue:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._stream_chat_completions())
        return self.output_queue, self.chat_history_queue

    async def close(self):
        self._task.cancel()

    async def _interrupt(self):
        while True:
            user_speaking = await self.interrupt_queue.get()
            if self._generating and user_speaking:
                self._task.cancel()
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                logger.info("Done cancelling LLM")
                self._generating = False
                self._task = asyncio.create_task(self._stream_chat_completions())

    async def set_interrupt(self, interrupt_queue: asyncio.Queue):
        self.interrupt_queue = interrupt_queue
        self._interrupt_task = asyncio.create_task(self._interrupt())



import asyncio
import os
import threading

import numpy as np
import torch

# Brute force torchaudio to use ffmpeg 7, otherwise it will use ffmpeg 6 which causes issues on mac
os.environ["TORIO_USE_FFMPEG_VERSION"] = "7"
# Although torchaudio is not utilized directly in this context, it is necessary to attempt its import as it is a dependency for Silero.
import torchaudio  # noqa: F401

from realtime.plugins.base_plugin import Plugin
from realtime.utils.cloneable_queue import CloneableQueue


class SileroVAD(Plugin):
    def __init__(
        self,
        buffer_duration: float = 0.1,
        audio_sample_rate: int = 8000,
        sensitivity_threshold: float = 0.91,
    ):
        if audio_sample_rate not in [8000, 16000]:
            raise ValueError("Silero VAD only supports 8KHz and 16KHz sample rates")

        torch.set_num_threads(1)

        (self.model, self.utils) = torch.hub.load(
            repo_or_dir="snakers4/silero-vad", model="silero_vad", force_reload=False
        )

        self.buffer_duration = buffer_duration
        self.audio_sample_rate = audio_sample_rate
        self.buffer_samples = int(self.audio_sample_rate * self.buffer_duration)

        self.sensitivity_threshold = sensitivity_threshold

        self._loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self.output_queue = CloneableQueue()
        self.user_speaking = False

    async def run(self, input_queue: CloneableQueue) -> CloneableQueue:
        self.input_queue = input_queue
        self._vad_thread = threading.Thread(target=self.execute_vad, daemon=True)
        self._vad_thread.start()
        return self.output_queue

    def execute_vad(self):
        try:
            buffer = None
            while True:
                future = asyncio.run_coroutine_threadsafe(self.input_queue.get(), self._loop)
                audio_frame = future.result()
                audio_data_int16 = audio_frame.to_ndarray()
                audio_data_float32 = self.convert_int_to_float(audio_data_int16)
                if buffer is None:
                    buffer = audio_data_float32
                else:
                    buffer = np.concatenate((buffer, audio_data_float32))

                if buffer.shape[0] > self.buffer_samples:
                    confidence_level = self.model(
                        torch.from_numpy(buffer[: self.buffer_samples]), self.audio_sample_rate
                    ).item()
                    buffer = buffer[-self.buffer_samples :]

                    is_speaking = confidence_level > self.sensitivity_threshold
                    if not is_speaking:
                        self.user_speaking = False
                    elif self._loop and is_speaking and not self.user_speaking:
                        print("silero", is_speaking, confidence_level)
                        self.user_speaking = True
                        asyncio.run_coroutine_threadsafe(self.output_queue.put(is_speaking), self._loop)
        except BaseException:
            # This is triggered by an empty audio buffer
            return False

    def convert_int_to_float(self, sound):
        try:
            maximum_absolute_value = np.abs(sound).max()
            sound = sound.astype("float32")
            if maximum_absolute_value > 0:
                sound *= 1 / 32768
            sound = sound.squeeze()  # Adjust based on specific needs
            return sound
        except ValueError:
            return sound



import asyncio
import base64
import io
import logging
import time
import wave

import av
import numpy as np
from fastapi import WebSocket

from realtime.streams import AudioStream, ByteStream, TextStream, VideoStream
from realtime.data import AudioData


class WebsocketInputStream:
    """
    Handles incoming WebSocket messages and streams audio and text data.

    Attributes:
        ws (WebSocket): The WebSocket connection.
    """

    def __init__(self, ws: WebSocket, sample_rate: int = 48000):
        self.ws = ws
        self.sample_rate = sample_rate

    async def run(self, audio_stream: AudioStream, message_stream: TextStream, video_stream: VideoStream):
        """
        Starts the task to process incoming WebSocket messages.

        Returns:
            Tuple[AudioStream, TextStream]: A tuple containing the audio and message streams.
        """
        self.audio_output_stream = audio_stream
        self.message_stream = message_stream

        # TODO: Implement video stream processing
        self.video_stream = video_stream

        audio_data = b""
        while True:
            try:
                data = await self.ws.receive_json()
                if data.get("type") == "message":
                    await self.message_stream.put(data.get("data"))
                elif data.get("type") == "audio":
                    audio_bytes = base64.b64decode(data.get("data"))
                    audio_data = AudioData(audio_bytes, sample_rate=self.sample_rate)
                    await self.audio_output_stream.put(audio_data)
            except Exception as e:
                logging.error("websocket: Exception", e)
                raise asyncio.CancelledError()


class WebsocketOutputStream:
    """
    Handles outgoing WebSocket messages by streaming audio and text data.

    Attributes:
        ws (WebSocket): The WebSocket connection.
    """

    def __init__(self, ws: WebSocket):
        self.ws = ws

    async def run(
        self, audio_stream: AudioStream, message_stream: TextStream, video_stream: VideoStream, byte_stream: ByteStream
    ):
        """
        Starts tasks to process and send byte and text streams.

        Args:
            audio_stream (AudioStream): The audio stream to send.
            message_stream (TextStream): The text stream to send.
            video_stream (VideoStream): The video stream to send.
            byte_stream (ByteStream): The byte stream to send.
        """
        # TODO: Implement video stream and audio stream processing
        await asyncio.gather(self.task(byte_stream), self.task(message_stream))

    async def task(self, input_stream):
        """
        Sends data from the input stream over the WebSocket.

        Args:
            input_stream (Stream): The stream from which to send data.
        """
        while True:
            data = await input_stream.get()
            if data is None:
                json_data = {"type": "audio_end", "timestamp": time.time()}
                await self.ws.send_json(json_data)
            elif isinstance(data, AudioData):
                data = data.get_bytes()
                output_bytes_io = io.BytesIO()
                in_memory_wav = wave.open(output_bytes_io, "wb")

                # TODO: Get channels, sample width, and sample rate from TTS module instead of hardcoding it
                in_memory_wav.setnchannels(1)
                in_memory_wav.setsampwidth(2)
                in_memory_wav.setframerate(16000)

                in_memory_wav.writeframes(data)
                output_bytes_io.seek(0)
                data = output_bytes_io.read()
                json_data = {"type": "audio", "data": base64.b64encode(data).decode(), "timestamp": time.time()}
                await self.ws.send_json(json_data)
            elif isinstance(data, str):
                json_data = {"type": "message", "data": data, "timestamp": time.time()}
                await self.ws.send_json(json_data)
            else:
                raise ValueError(f"Unsupported data type: {type(data)}")



import asyncio
import json
import logging
import time

import google.generativeai as genai
import PIL.PngImagePlugin  # Not used but needed to make Gemini API work with PIL  # noqa: F401
from typing import Optional

from realtime.plugins.vision_plugin import VisionPlugin
from realtime.streams import TextStream, VideoStream

logger = logging.getLogger(__name__)


class GeminiVision(VisionPlugin):
    def __init__(
        self,
        model: str = "gemini-1.5-flash-latest",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        system_prompt: Optional[str] = None,
        auto_respond: Optional[int] = None,
        temperature: float = 1.0,
        wait_for_first_user_response: bool = False,
        chat_history: bool = True,
    ):
        super().__init__()
        self._model: str = model
        self._client = genai.GenerativeModel(model)
        self._history = []
        self._system_prompt = system_prompt
        self._temperature = temperature
        if self._system_prompt is not None:
            self._history.append({"role": "user", "parts": [self._system_prompt]})
        self._auto_respond = auto_respond
        self.wait_for_first_user_response = wait_for_first_user_response
        self._time_last_response = None
        self.chat_history_queue = TextStream()
        self._chat_history = chat_history
        self._safety_settings = [
            {
                "category": "HARM_CATEGORY_HARASSMENT",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_HATE_SPEECH",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
                "threshold": "BLOCK_NONE",
            },
        ]

    async def _stream_chat_completions(self):
        while True:
            if self._auto_respond:
                try:
                    prompt = await asyncio.wait_for(self.text_input_queue.get(), timeout=0.2)
                except asyncio.TimeoutError:
                    if (
                        self._time_last_response is not None
                        and time.time() - self._time_last_response < self._auto_respond
                    ):
                        continue
                    prompt = ""
            else:
                prompt = await self.text_input_queue.get()
            if prompt is None:
                continue
            if prompt == "" and self.image_input_queue.qsize() == 0:
                continue
            self._generating = True
            start_time = time.time()
            self._history.append(
                {
                    "role": "user",
                    "parts": [],
                }
            )
            if prompt != "":
                self._history[-1]["parts"].append(prompt)
                self.chat_history_queue.put_nowait(
                    json.dumps(
                        {
                            "role": "user",
                            "content": prompt,
                        }
                    )
                )
            if self.image_input_queue.qsize() > 0:
                while self.image_input_queue.qsize() > 0:
                    image = self.image_input_queue.get_nowait()
                self._history[-1]["parts"].append(image[0])
                logger.info("Google AI image %s", image[1])

            try:
                response = await self._client.generate_content_async(
                    self._history,
                    stream=True,
                    generation_config=genai.types.GenerationConfig(max_output_tokens=75, temperature=self._temperature),
                    safety_settings=self._safety_settings,
                )
            except Exception as e:
                logger.error("Google AI vision timeout %s", e)
                self._history.pop()
                continue

            # if prompt != "":
            #     self._history[-1]["parts"] = self._history[-1]["parts"][:1]
            # else:
            if not self._chat_history:
                self._history.pop()
            self._history.append({"role": "model", "parts": [""]})
            logger.info("Google AI LLM TTFB: %s", time.time() - start_time)
            async for chunk in response:
                if chunk:
                    try:
                        text = chunk.text
                    except:
                        continue
                    self._history[-1]["parts"][0] += text
                    await self.output_queue.put(text)
            logger.info("llm %s", self._history[-1]["parts"][0])
            self.chat_history_queue.put_nowait(
                json.dumps(
                    {
                        "role": "assistant",
                        "content": self._history[-1]["parts"][0],
                    }
                )
            )
            if not self._chat_history:
                self._history.pop()
            self._time_last_response = time.time()
            self._generating = False
            await self.output_queue.put(None)

    async def run(self, text_input_queue: TextStream, image_input_queue: VideoStream) -> TextStream:
        self.text_input_queue = text_input_queue
        self.image_input_queue = image_input_queue
        self._tasks = [asyncio.create_task(self._stream_chat_completions())]
        return self.output_queue, self.chat_history_queue



import sys

if sys.version_info[:2] < (3, 9):
    raise RuntimeError("This version of Realtime requires at least Python 3.9")
if sys.version_info[:2] >= (3, 13):
    raise RuntimeError("This version of Realtime does not support Python 3.13+")

import sentry_sdk

sentry_sdk.init(
    dsn="https://8842715aaa1b7fd845f8a55eea150394@o4506805333983232.ingest.us.sentry.io/4507603326795776",
    # Set traces_sample_rate to 1.0 to capture 100%
    # of transactions for performance monitoring.
    traces_sample_rate=1.0,
    # Set profiles_sample_rate to 1.0 to profile 100%
    # of sampled transactions.
    # We recommend adjusting this value in production.
    profiles_sample_rate=1.0,
)

try:
    from .app import App  # noqa: F401
    from .streaming_endpoint import streaming_endpoint  # noqa: F401
    from .web_endpoint import web_endpoint  # noqa: F401
    from .websocket import websocket  # noqa: F401
    from .data import AudioData, ImageData, TextData  # noqa: F401
    from .plugins.cartesia_tts import CartesiaTTS  # noqa: F401
    from .plugins.deepgram_stt import DeepgramSTT  # noqa: F401
    from .plugins.groq_llm import GroqLLM  # noqa: F401
    from .plugins.token_aggregator import TokenAggregator  # noqa: F401
    from .streams import AudioStream, VideoStream, TextStream  # noqa: F401
    from .ops.map import map  # noqa: F401
    from .ops.merge import merge  # noqa: F401
    from .plugins.azure_tts import AzureTTS  # noqa: F401
    from .plugins.fireworks_llm import FireworksLLM  # noqa: F401
    from .plugins.eleven_labs_tts import ElevenLabsTTS  # noqa: F401
except Exception:
    print()
    print("#" * 50)
    print("#" + "Something with the Realtime installation seems broken.".center(48) + "#")
    print("#" * 50)
    print()
    raise

__all__ = [
    "streaming_endpoint",
    "App",
    "web_endpoint",
    "websocket",
    "AudioData",
    "ImageData",
    "TextData",
    "CartesiaTTS",
    "DeepgramSTT",
    "GroqLLM",
    "TokenAggregator",
    "map",
    "merge",
    "AzureTTS",
    "ElevenLabsTTS",
    "FireworksLLM",
]



import asyncio
import functools
import logging
from realtime.server import RealtimeServer
from fastapi import WebSocket
from realtime.streams import AudioStream, TextStream, VideoStream, ByteStream
import inspect
from realtime.websocket.processors import WebsocketInputStream, WebsocketOutputStream


logger = logging.getLogger(__name__)


def websocket(path: str = "/"):
    """
    Decorator for handling WebSocket connections.
    TODO: Add video support
    """

    def decorator(func):
        async def websocket_handler(websocket: WebSocket):
            RealtimeServer().add_connection()
            try:
                await websocket.accept()
                audio_metadata = await websocket.receive_json()

                audio_input_q = None
                video_input_q = None
                text_input_q = None
                kwargs = {}

                signature = inspect.signature(func)
                parameters = signature.parameters
                for name, param in parameters.items():
                    if param.annotation == AudioStream:
                        audio_input_q = AudioStream(sample_rate=audio_metadata.get("sampleRate", 48000))
                        kwargs[name] = audio_input_q
                    elif param.annotation == VideoStream:
                        video_input_q = VideoStream()
                        kwargs[name] = video_input_q
                    elif param.annotation == TextStream:
                        text_input_q = TextStream()
                        kwargs[name] = text_input_q

                output_streams = await func(**kwargs)

                if not isinstance(output_streams, (list, tuple)):
                    output_streams = (output_streams,)

                aq, vq, tq = None, None, None
                for s in output_streams:
                    if isinstance(s, AudioStream):
                        aq = s
                    elif isinstance(s, VideoStream):
                        vq = s
                    elif isinstance(s, TextStream):
                        tq = s
                    elif isinstance(s, ByteStream):
                        bq = s

                # TODO: Update the default sample rate to be consistent across all plugins
                tasks = [
                    WebsocketInputStream(websocket, audio_metadata.get("sampleRate", 48000)).run(
                        audio_stream=audio_input_q, message_stream=text_input_q, video_stream=video_input_q
                    ),
                    WebsocketOutputStream(websocket).run(
                        audio_stream=aq, message_stream=tq, video_stream=vq, byte_stream=bq
                    ),
                ]

                await asyncio.gather(*tasks)
            except asyncio.CancelledError:
                logging.error("websocket: CancelledError")
            except Exception as e:
                logging.error("websocket: Error in websocket: ", e)
            finally:
                logging.info("websocket: Removing connection")
                RealtimeServer().remove_connection()

        fastapi_app = RealtimeServer().get_app()
        fastapi_app.websocket(path)(websocket_handler)

    return decorator



import traceback  # Add this import at the top of your file  # noqa: I001

import asyncio
import json
import logging
import time

from openai import AsyncOpenAI
from typing import Optional

from realtime.plugins.vision_plugin import VisionPlugin
from realtime.streams import TextStream, VideoStream

logger = logging.getLogger(__name__)


class OpenAIVision(VisionPlugin):
    def __init__(
        self,
        model: str = "gpt-4-turbo",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        system_prompt: Optional[str] = None,
        auto_respond: Optional[int] = None,
        temperature: float = 1.0,
        wait_for_first_user_response: bool = False,
        key_frame_threshold: float = 0.4,
    ):
        super().__init__()
        self._model: str = model
        self._client = AsyncOpenAI(api_key=api_key, base_url=base_url)
        self._history = []
        self._system_prompt = system_prompt
        self._temperature = temperature
        if self._system_prompt is not None:
            self._history.append({"role": "system", "content": self._system_prompt})
        self._auto_respond = auto_respond
        self.wait_for_first_user_response = wait_for_first_user_response
        self._time_last_response = None
        self.chat_history_queue = TextStream()
        self._key_frame_threshold = key_frame_threshold

    async def _stream_chat_completions(self):
        while True:
            if self._auto_respond:
                try:
                    prompt = await asyncio.wait_for(self.text_input_queue.get(), timeout=0.2)
                except asyncio.TimeoutError:
                    if (
                        self._time_last_response is not None
                        and time.time() - self._time_last_response < self._auto_respond
                    ):
                        continue
                    prompt = ""
            else:
                prompt = await self.text_input_queue.get()
            if prompt is None:
                continue
            if prompt == "" and len(self.video_frames_stack) == 0:
                continue
            self._generating = True
            start_time = time.time()
            self._history.append(
                {
                    "role": "user",
                    "content": [],
                }
            )
            if prompt != "":
                self._history[-1]["content"].append({"type": "text", "text": prompt})
                self.chat_history_queue.put_nowait(
                    json.dumps(
                        {
                            "role": "user",
                            "content": prompt,
                        }
                    )
                )
            if len(self.video_frames_stack) > 0:
                image = self.video_frames_stack.pop()
                self._history[-1]["content"].append({"type": "image_url", "image_url": {"url": image[0]}})
                logger.info("open ai image %s", image[1])

            try:
                chunk_stream = await asyncio.wait_for(
                    self._client.chat.completions.create(
                        model=self._model,
                        stream=True,
                        messages=self._history,
                        max_tokens=75,
                        temperature=self._temperature,
                    ),
                    timeout=4,
                )
            except:
                traceback.print_exc()
                logger.error("OpenAI vision timeout")
                self._history.pop()
                continue

            if prompt != "":
                self._history[-1]["content"] = self._history[-1]["content"][:1]
            else:
                self._history[-1]["content"] = []
            self._history.append({"role": "assistant", "content": ""})
            logger.info("OpenAI LLM TTFB: %s", time.time() - start_time)
            async for chunk in chunk_stream:
                if len(chunk.choices) == 0:
                    continue

                elif chunk.choices[0].delta.content:
                    self._history[-1]["content"] += chunk.choices[0].delta.content
                    await self.output_queue.put(chunk.choices[0].delta.content)
            logger.info("llm %s", self._history[-1]["content"])
            self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
            self._time_last_response = time.time()
            self._generating = False
            await self.output_queue.put(None)

    async def run(self, text_input_queue: TextStream, image_input_queue: VideoStream) -> TextStream:
        self.text_input_queue = text_input_queue
        self.image_input_queue = image_input_queue
        self._tasks = [asyncio.create_task(self._stream_chat_completions()), asyncio.create_task(self.process_video())]
        return self.output_queue, self.chat_history_queue



import asyncio
import base64
import json
import logging
import os
import uuid
from typing import Optional
from urllib.parse import urlencode

import websockets

from realtime.data import AudioData
from realtime.plugins.base_plugin import Plugin
from realtime.streams import ByteStream, TextStream, AudioStream
from realtime.utils import tracing


class CartesiaTTS(Plugin):
    """
    A plugin for text-to-speech synthesis using the Cartesia API.

    This class handles the connection to the Cartesia TTS service, sends text for synthesis,
    and receives the generated audio data.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        voice_id: str = "a0e99841-438c-4a64-b679-ae501e7d6091",
        model: str = "sonic-english",
        output_encoding: str = "pcm_s16le",
        output_sample_rate: int = 16000,
        stream: bool = True,
        base_url: str = "wss://api.cartesia.ai/tts/websocket",
        cartesia_version: str = "2024-06-10",
    ):
        """
        Initialize the CartesiaTTS plugin.

        Args:
            api_key (Optional[str]): The API key for Cartesia. If not provided, it will be read from the CARTESIA_API_KEY environment variable.
            voice_id (str): The ID of the voice to use for synthesis.
            model (str): The model to use for synthesis.
            output_encoding (str): The encoding of the output audio.
            output_sample_rate (int): The sample rate of the output audio.
            stream (bool): Whether to stream the audio output.
            base_url (str): The base URL for the Cartesia API.
            cartesia_version (str): The version of the Cartesia API to use.
        """
        super().__init__()

        self._generating: bool = False
        self._task: Optional[asyncio.Task] = None
        self._ws: Optional[websockets.WebSocketClientProtocol] = None
        self._interrupt_task: Optional[asyncio.Task] = None

        # Set up API key
        self.api_key: str = api_key or os.environ.get("CARTESIA_API_KEY")
        if self.api_key is None:
            raise ValueError("Cartesia API key is required")

        # Set up TTS parameters
        self.voice_id: str = voice_id
        self.model: str = model
        self.output_encoding: str = output_encoding
        self.output_sample_rate: int = output_sample_rate
        self.output_queue: AudioStream = AudioStream()
        self.stream: bool = stream
        self.base_url: str = base_url
        self.cartesia_version: str = cartesia_version
        self._current_context_id: Optional[str] = None
        self._ws = None

        # Initialize queues
        self.input_queue: Optional[TextStream] = None
        self.interrupt_queue: Optional[asyncio.Queue] = None

    def run(self, input_queue: TextStream) -> ByteStream:
        """
        Start the TTS synthesis process.

        Args:
            input_queue (TextStream): The input queue containing text to synthesize.

        Returns:
            ByteStream: The output queue containing synthesized audio data.
        """
        self.input_queue = input_queue
        self._task = asyncio.create_task(self.synthesize_speech())
        return self.output_queue

    async def connect_websocket(self):
        query_params = {
            "cartesia_version": self.cartesia_version,
            "api_key": self.api_key,
        }
        try:
            self._ws = await websockets.connect(f"{self.base_url}?{urlencode(query_params)}")
        except Exception as e:
            logging.error("Error connecting to Cartesia TTS: %s", e)
            raise asyncio.CancelledError()

    async def synthesize_speech(self):
        """
        Main method for speech synthesis. Connects to the Cartesia API and manages
        the send_text and receive_audio coroutines.
        """

        async def send_text():
            """Send text chunks to the Cartesia API for synthesis."""
            first_chunk = True
            try:
                while True:
                    text_chunk = await self.input_queue.get()
                    if first_chunk:
                        await self.connect_websocket()
                        first_chunk = False
                    if text_chunk is None or text_chunk == "":
                        if self._current_context_id is None:
                            continue
                        payload = {
                            "transcript": "",
                            "context_id": self._current_context_id,
                            "model_id": self.model,
                            "continue": False,
                            "voice": {"mode": "id", "id": self.voice_id},
                            "output_format": {
                                "encoding": self.output_encoding,
                                "sample_rate": self.output_sample_rate,
                                "container": "raw",
                            },
                        }
                        await self._ws.send(json.dumps(payload))

                        self._current_context_id = None
                        continue
                    if self._current_context_id is None:
                        if self._generating:
                            continue
                        self._current_context_id = str(uuid.uuid4())
                    tracing.register_event(tracing.Event.TTS_START)
                    logging.info("Generating TTS: %s", text_chunk)
                    payload = {
                        "voice": {"mode": "id", "id": self.voice_id},
                        "output_format": {
                            "encoding": self.output_encoding,
                            "sample_rate": self.output_sample_rate,
                            "container": "raw",
                        },
                        "transcript": text_chunk,
                        "model_id": self.model,
                        "context_id": self._current_context_id,
                        "continue": True,
                    }
                    self._generating = True
                    await self._ws.send(json.dumps(payload))
            except Exception as e:
                logging.error("Error sending text to Cartesia TTS: %s", e)
                self._generating = False
                self._current_context_id = None
                await self.output_queue.put(None)
                raise asyncio.CancelledError()

        async def receive_audio():
            """Receive synthesized audio from the Cartesia API and put it in the output queue."""
            try:
                total_audio_bytes = 0
                is_first_chunk = True
                while True:
                    if self._ws is None:
                        await asyncio.sleep(0.2)
                        continue
                    response = await self._ws.recv()
                    response = json.loads(response)
                    if response["type"] == "chunk":
                        audio_bytes = base64.b64decode(response["data"])
                        total_audio_bytes += len(audio_bytes)
                        if is_first_chunk:
                            tracing.register_event(tracing.Event.TTS_TTFB)
                            is_first_chunk = False
                        await self.output_queue.put(
                            AudioData(
                                audio_bytes,
                                sample_rate=self.output_sample_rate,
                            )
                        )
                    elif response["type"] == "done":
                        tracing.register_event(tracing.Event.TTS_END)
                        tracing.register_metric(tracing.Metric.TTS_TOTAL_BYTES, total_audio_bytes)
                        total_audio_bytes = 0
                        self._generating = False
                        is_first_chunk = True
                        tracing.log_timeline()
                        await self.output_queue.put(None)
                    else:
                        logging.error("Unknown response type in Cartesia TTS: %s", response)
            except Exception as e:
                logging.error("Error receiving audio from Cartesia TTS: %s", e)
                self._generating = False
                self._current_context_id = None
                await self.output_queue.put(None)
                raise asyncio.CancelledError()

        try:
            await asyncio.gather(send_text(), receive_audio())
        except asyncio.CancelledError:
            logging.info("TTS cancelled")
            self._generating = False
            self._current_context_id = None

    async def close(self):
        """Close the websocket connection and cancel the main task."""
        if self._ws:
            await self._ws.close()
        if self._task:
            self._task.cancel()

    async def _interrupt(self):
        """
        Handle interruptions (e.g., when the user starts speaking).
        Cancels ongoing TTS generation and clears the output queue.
        """
        while True:
            user_speaking = await self.interrupt_queue.get()
            if self._generating and user_speaking:
                if self._task:
                    self._task.cancel()
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                logging.info("Done cancelling TTS")
                self._generating = False
                self._task = asyncio.create_task(self.synthesize_speech())

    async def set_interrupt(self, interrupt_queue: asyncio.Queue):
        """
        Set up the interrupt queue and start the interrupt handling task.

        Args:
            interrupt_queue (asyncio.Queue): The queue for receiving interrupt signals.
        """
        self.interrupt_queue = interrupt_queue
        self._interrupt_task = asyncio.create_task(self._interrupt())
